/*
 * initvfork.c 
 */

//
#define O_ACCMODE         0003
#define O_RDONLY            00
#define O_WRONLY            01
#define O_RDWR              02
#define O_CREAT           0100  /* not fcntl */
#define O_EXCL            0200  /* not fcntl */
#define O_NOCTTY          0400  /* not fcntl */
#define O_TRUNC          01000  /* not fcntl */
#define O_APPEND         02000
#define O_NONBLOCK       04000
#define O_NDELAY        O_NONBLOCK
#define O_SYNC          010000
#define O_FSYNC         O_SYNC
#define O_ASYNC         020000

//
#define MS_RDONLY    1 /* mount read-only */
#define MS_NOSUID    2 /* ignore suid and sgid bits */
#define MS_NODEV     4 /* disallow access to device special files */
#define MS_NOEXEC    8 /* disallow program execution */
#define MS_SYNC     16 /* writes are synced at once */
#define MS_REMOUNT  32 /* alter flags of a mounted FS */
#define MS_MGC_VAL 0xC0ED0000

//include/linux/wait.h
#define WNOHANG         0x00000001
#define WUNTRACED       0x00000002
#define __WCLONE        0x80000000

//include/asm/ioctl.h
#include <asm/ioctl.h>

//include/asm/ioctls.h
#define TIOCSCTTY       _IO('T', 14) /* 0x540E */
#define TIOCSPGRP       _IOW('t', 118, int)
#define TIOCGSID        _IOR('T', 41, pid_t) /* 0x5429 */ /* Return the session ID of FD */

//include/asm/signal.h
#define SIGCHLD         17
//#define SIG_IGN ((__sighandler_t)1)     /* ignore signal */
#define SIG_IGN		1     /* ignore signal */


#include "nlc.h"
extern int errno;
#define PERROR(str) do { char *s = str; write(2,s,strlen(s)); } while(0)
#define INTERACTIVE


char	*envp[] = { "PATH=/bin:/sbin:/usr/bin", 0 };


int vfork_execve(char *file, char **argv, char **envp)
{
	int	cpid, pid;
	  /**/
	if((cpid = vfork()) == 0){
//write(1,"[vfork_execve-0]",16);
		pid = getpid();
		setsid();
		setpgid(pid,pid);
		ioctl(0,TIOCSCTTY,0);
		ioctl(0,TIOCSPGRP,&pid);
//write(1,"[vfork_execve-1]",16);
		execve(file,argv,envp);
//write(1,"[vfork_execve-2]",16);
		exit(1);
//write(1,"[vfork_execve-3]",16);
	}
	return cpid;
}


int do_exec(char **cmd)
{
	int	pid;
	//int	status;
	char	*p, *e, b[512];
	  /**/
	for(p = cmd[0]; *p != 0; ++p){
		if(*p == '/'){
			break;
		}
	}
	if(*p == '/'){
		p = cmd[0];
	}else{
		//putss("cmd[0]=",cmd[0]);
		e = b + sizeof(b);
		p = prstr(b,e,"/sbin/");
		p = prstr(p,e,cmd[0]);
		if(p >= e){
			return -1;
		}
		*p = 0;
		p = b;
		//putss("file=", b);
	}
	pid = vfork_execve(p,cmd,envp);
	//putsn("pid=",pid);
	ioctl(0,TIOCSPGRP,&pid);
	//write(1, "\nWaiting ...", 11);
	//pid = waitpid(pid,&status,WUNTRACED);
	//putsn("pid=",pid);
	//putsx("status=",status);
	return pid;
}


void handler(int signum)
{
	int	pid, status;
	  /**/
	putsn("handler:",signum);
	if(signum == 17){
		pid = waitpid(-1,&status,3/*WNOHANG|WUNTRACED*/);
		putsn("SIGCHLD: pid=",pid);
		putsx("SIGCHLD: status=",status);
	}
}


char *str2arg(char *b, char *e, char **avb, char **ave, int *acp, int *m)
{
	int	c, cond;
	char	*p, *q, **avp;
	  /**/
	*m = 0;
	for(p = q = b, avp = avb, cond = 0;;){
		if(p >= e){
			p = 0;
			break;
		}
		c = *(unsigned char*)p++;
		if(c <= 32 || c == 127 || (c & 128) != 0 || c == '&' || c == '?'){
			if(cond == 1){
				*q++ = 0;
				++avp;
				cond = 0;
			}else if(cond == 3){
				--avp;
				cond = 2;
			}
			if(c == '&' || c == '?'){
				*m = c;
				break;
			}
			if(c == 0 || c == '\n'){
				break;
			}
		}else{
			if(cond == 0){
				if(avp < ave){
					*avp = q;
					*q++ = c;
					cond = 1;
				}else{
					avp = avb - ave + avb;
					cond = 3;
				}
			}else if(cond == 1){
				*q++ = c;
			}else if(cond == 2){
				cond = 3;
			}
		}
	}
	*acp = avp - avb;
	return p;
}


int main(int ac, char **av)
{
	int	i, r, fd, cont, m, status, pid;
	char	*p, *q, *b, *e, *s;
	char	*file;
	char	*prompt_e;
	char	prompt[8];
	char	*argv[32], linebuf[1024];
	  /**/
	putstr(__FILE__"\n");
//
	pid = getpid();
//		putsn("getpid()=",pid);
	r = setsid();
//		putsn("setsid()=",r);
	r = setpgid(0,0);
//		putsn("setsid()=",r);
#if 0
	ioctl(1,TIOCGSID,&i);
	writeln_sd(1,buf,buf+sizeof(buf),"SID: ", i);
#endif
	r = ioctl(1,TIOCSPGRP,&pid);
	i = errno;
//		putsn("ioctl(TIOCSPGRP)=",r);
	if(r < 0){
//		putsn("  --> errno=",i);
	}
	fd = open("/dev/tty",O_RDWR|O_NONBLOCK);
	i = errno;
//		putsn("open(/dev/tty)=",fd);
	if(fd >= 0){
		close(fd);
	}else{
//		putsn("  --> errno=",i);
	}
	r = ioctl(1,TIOCSCTTY,0);
	if(i < 0){
//		putsn("TIOCSCTTY --> errno=",i);
	}
	signal(SIGCHLD,SIG_IGN);
//
	// mount /dev/hda / ext2 rw MS_MGC_VAL|MS_REMOUNT 0
	r = mount("/dev/hda","/",0,MS_MGC_VAL|MS_REMOUNT,0);
	//putsn("mount: ret=",r);
	if(r < 0){
		putsn("mount: errno=",errno);
		PERROR("/dev/hda");
		exit(1);
	}
//
#if 0
	p = linebuf;
	e = linebuf + sizeof(linebuf);
	p = prstr(p,e,"signal");
	for(i = 0; i < 32; ++i){
		r = signal(i/*SIGCHLD*/,handler);
		p = prstr(p,e," ");
		p = prnum(p,e,i,0);
		p = prstr(p,e,"=");
		p = prnum(p,e,r,0);
	}
	p = prstr(p,e,"\n");
	write(1,linebuf,p-linebuf);
	//signal(17/*SIGCHLD*/,handler);
#endif
	fd = 0;
	if(ac > 1){
		fd = open(file="/sbin/rc",0,0);
//--		putsn("open: fd=",fd);
		if(fd < 0){
			putsn("open: errno=",errno);
			PERROR(file);
			exit(1);
		}
	}
	p = prompt;
#if 0
	if(pid >= 10){
		*p++ = pid / 10 + '0';
	}
	*p++ = pid % 10 + '0';
#endif
	*p++ = '$';
	*p++ = ' ';
	prompt_e = p;
	*p++ = 0;
	for(b = linebuf, e = linebuf + sizeof(linebuf), cont = 0;;){
		if(fd == 0){
			write(2,prompt,prompt_e-prompt);
		}
		for(;;){
			r = read(fd,b,e-b);
			if(fd != 0 || r >= 0){
				break;
			}
		}
//--		putsn("read: ret=",r);
		if(r < 0){
			putsn("read: errno=",errno);
			PERROR("read");
			break;
		}else if(r == 0){
			break;
		}
//--		write(2,linebuf,r);
		for(p = linebuf, q = b + r; p < q;){
			p = str2arg(s = p, q, argv, argv + sizeof(argv)/sizeof(argv[0]), &r, &m);
			if(p == 0){
				if(s == linebuf){
					b = linebuf;
					puts("too long line. ");
					cont = 1;
				}else{
					memcpy(linebuf,s, r = q - s);
					b = linebuf + r;
				}
				break;
			}
			if(cont != 0){
				cont = 0;
				continue;
			}
			if(r < 0){
				putsn("too meny args. ",-r);
				continue;
			}
			if(r == 0 || argv[0][0] == '#'){
				continue;
			}
			if(strcmp(argv[0],"cd") == 0){
				if(r == 2){
//--					prarg(r-1,argv+1);
					chdir(argv[1]);
				}
				continue;
			}
			argv[r] = 0;
			i = do_exec(argv);
			if(m != '&' && i >= 0){
				pid = waitpid(i,&status,WUNTRACED);
				// putsn("pid=",pid);
				// putsx("status=",status);
#if 0
				if(!WIFEXITED(status)){
					if(WIFSTOPPED(status)){
						i = WSTOPSIG(status);
					}else{
						i = WTERMSIG(status);
					}
					putsn("signal=",i);
					putsn("WCOREDUMP=",WCOREDUMP(status));
				}
#endif
			}
		}
	}
	close(fd);
	return;
}
